#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "strsubs.h"
#include "symsubs.h"

// Next expected token
#define EXPECT_ANYTHING 0
#define EXPECT_IF_OR_OP 1
#define EXPECT_OP       2
#define EXPECT_DST      3
#define EXPECT_DST1     4
#define EXPECT_DST2     5
#define EXPECT_DONE     7
#define EXPECT_DST2P    10
#define EXPECT_DST1P    11
#define EXPECT_UNKNOWN  13
#define EXPECT_OP2A     14
#define EXPECT_OP2B     15
#define EXPECT_OP2C     16
#define EXPECT_OP1B     18
#define EXPECT_OP1C     19
#define EXPECT_OP3AX    23
#define EXPECT_OP2PX    24
#define EXPECT_OP2AX    25
#define EXPECT_OP2EX    26
#define EXPECT_OP2CX    27
#define EXPECT_OP2DX    28
#define EXPECT_OP1AX    29
#define EXPECT_OP0      30
#define EXPECT_OP2XX    31
#define EXPECT_REPXX    32

// Printing modes
#define PRINT_NOCODE    0
#define PRINT_CODE      1
#define PRINT_NONE      2

// Spin section types
#define MODE_DAT 0
#define MODE_CON 1

// C, Z and I bits
#define C_BIT (1 << 20)
#define Z_BIT (1 << 19)
#define I_BIT (1 << 18)

// Keywords and delimiters
char *SectionKeywords[] = {"dat", "con", "pub", "pri", "var", 0};

FILE *infile;
FILE *lstfile;
FILE *binfile;

SymbolT SymbolTable[MAX_SYMBOLS];
char databuffer[1000];

int numsym = 0;
int cog_addr = 0;
int hub_addr = 0;
int printlevel = 3;
int unicode = 0;
int last_size = 1;
int hubmode = 0;
int bincount = 0;
char buffer2[300];

void DumpIt(int printflag, void *ptr, int num);
void PrintIt(int printflag, int hub_addr, int cog_addr, int data_size, char *buffer2, void *ptr);

void PrintError(char *str, void *parm1, void *parm2)
{
    printf(str, parm1, parm2);
    printf("%s\n", buffer2);
    fprintf(lstfile, str, parm1, parm2);
}

void PrintUnexpected(int num, char *str)
{
    PrintError("ERROR: (%d) Unexpected symbol \"%s\"\n", (void *)num, str);
}

int CheckExpected(char *ptr1, int i, char **tokens, int num)
{
    if (i >= num)
    {
        PrintError("ERROR: Expected \"%s\", but found EOL\n", ptr1, 0);
        return 1;
    }
    if (strcmp(ptr1, tokens[i]))
    {
        PrintError("ERROR: Expected \"%s\", but found \"%s\"\n", ptr1, tokens[i]);
        return 1;
    }
    return 0;
}

// This routine handles the P2 pointer syntax
int EncodePointerField(int *pindex, char **tokens, int num)
{
    int value;
    int index = 0;
    int negate = 0;
    int i = *pindex;
    int is_float = -1;
    int retval = 0x100;

    // Check for pre-increment/decrement
    if (!strcmp(tokens[i], "++"))
    {
        retval |= 0x41;
        if (++i >= num) { *pindex = i; return retval; }
        //value = StrCompNoCase(tokens[i], "ptrb");
    }
    else if (!strcmp(tokens[i], "--"))
    {
        negate = 1;
        retval |= 0x5f;
        if (++i >= num) { *pindex = i; return retval; }
        //value = StrCompNoCase(tokens[i], "ptrb");
    }

    value = StrCompNoCase(tokens[i], "ptrb");
    retval |= (value & 1) << 7;
    if (i >= num - 1) { *pindex = i; return retval; }

    // Check for post-increment/decrement
    if (!strcmp(tokens[i+1], "++"))
    {
        retval |= 0x61;
        if (++i >= num - 1) { *pindex = i; return retval; }
    }
    else if (!strcmp(tokens[i+1], "--"))
    {
        negate = 1;
        retval |= 0x7f;
        if (++i >= num - 1) { *pindex = i; return retval; }
    }

    // Check for opening bracket
    if (strcmp(tokens[i+1], "[")) { *pindex = i; return retval; }

    if (++i > num - 1)
    {
	*pindex = i;
	return -1;
    }

    i++;
    EvaluateExpression(12, &i, tokens, num, &value, &is_float);
    if (negate) value = -value;

    if (++i >= num)
    {
	*pindex = i;
	return -1;
    }

    if (value < -16 || value > 15)
    {
        fprintf(lstfile, "Warning: pointer index %d is out of bounds\n", value);
    }

    retval = (retval & ~31) | (value & 31);

    if (i >= num)
    {
        *pindex = i;
	return -1;
    }

    if (CheckExpected("]", i, tokens, num))
    {
        *pindex = i;
        return -1;
    }

    *pindex = i;
    return retval;
}

void GenerateAugx(int opcode, int value, int dfield)
{
    opcode &= 0xf0000000; // Extract the condition code
    if (dfield)
        opcode |= 0x0f800000; // AUGD
    else
        opcode |= 0x0f000000; // AUGS
    opcode |= (value >> 9) & 0x7fffff;
    PrintIt(PRINT_CODE, hub_addr, cog_addr, 4, buffer2, &opcode);
    DumpIt(PRINT_CODE, &opcode, 4);
    buffer2[0] = 0;
    cog_addr += 4;
    hub_addr += 4;
}

// This routine processes the source or destination field
// P2 pointer expressions are allowed if type is >= 2
// Immediate values using a # are allowed if type is >= 1
// The value is returned if successful, and bit at 0x200 is set if immediate
// A minus one is return if unsuccessful
int EncodeAddressField(int *pindex, char **tokens, int num, int type, int opcode, int dfield)
{
    int retval = 0;
    int i = *pindex;
    int index, value, errnum;
    char *name = tokens[i];
    int extended = 0;
    int is_float = -1;

    if (StrCompNoCase(name, "ptra") || StrCompNoCase(name, "ptrb") ||
        !strcmp(name, "++") || !strcmp(name, "--"))
    {
	if (type >= 2)
	{
            value = EncodePointerField(&i, tokens, num);
	    if (value >= 0) value |= 0x200;
	}
	else
	{
	    PrintError("ERROR: PTR not allowed\n", 0, 0);
	    value = -1;
	}
	*pindex = i;
	return value;
    }

    if (!strcmp(name, "#") || !strcmp(name, "##"))
    {
        if (!strcmp(name, "##")) extended = 1;
	if (type < 1)
	{
            PrintError("ERROR: Bad symbol type - %d\n", (void *)type, 0);
	    *pindex = i;
	    return -1;
	}
	retval = 0x200;
	if (++i >= num)
	{
            PrintError("ERROR: Unexpected EOL\n", 0, 0);
	    *pindex = i;
	    return -1;
	}
        errnum = EvaluateExpression(12, &i, tokens, num, &value, &is_float);
        if (extended) GenerateAugx(opcode, value, dfield);
    }
    else
    {
        errnum = EvaluateExpression(12, &i, tokens, num, &value, &is_float);
    }

    if (errnum)
    {
        *pindex = i;
        return -1;
    }
    retval |= (value & 511);
    *pindex = i;
    return retval;
}

// Handle conditional flags, such as wc or nr
int ProcessWx(int *pi, char **tokens, int num, int *popcode, int *pexpect)
{
    int index;
    SymbolT *s;
    int i = *pi;

    while (++i < num)
    {
        index = FindSymbol(tokens[i]);
        if (index < 0) { *pi = i; return 1; }
        s = &SymbolTable[index];

        if (s->type != TYPE_WX && s->type != TYPE_NX)
        {
            PrintUnexpected(6, tokens[i]);
            *pi = i;
            return 1;
        }
        if (s->type == TYPE_WX)
        {
            *popcode |= s->value;
        }
        else
            *popcode &= ~s->value;
        if (++i >= num) break;
        if (CheckExpected(",", i, tokens, num))
        {
            *pi = i;
            return 1;
        }
    }

    // Check if done
    if (i < num)
    {
        *pi = i;
        PrintUnexpected(9, tokens[i]);
        return 1;
    }

    *pi = i;
    *pexpect = EXPECT_DONE;

    return 0;
}

// Handle large address fields
int ProcessBigSrc(int *pi, char **tokens, int num, int *popcode, int *pexpect)
{
    int value;
    int rflag = 0;
    int is_float = -1;
    int target_hubmode;

    (*pi)++;
    if (EvaluateExpression(12, pi, tokens, num, &value, &is_float)) return 1;
    target_hubmode = value >= 0x400;
    if (hubmode == target_hubmode) rflag = 1;
    if (rflag)
    {
        *popcode |= C_BIT;
        if (hubmode)
            value -= hub_addr + 4;
        else
            value = (value << 2) - cog_addr - 4;
    }
    *popcode |= (value & 0xfffff);
    return ProcessWx(pi, tokens, num, popcode, pexpect);
}

// This routine returns the next expected state based on the op_type
int GetExpect(int op_type, int i, int num)
{
    int expect;

    if (op_type == TYPE_OP0)
    {
	if (i != num - 1)
            expect = EXPECT_OP0;
	else
            expect = EXPECT_DONE;
    }
    else if (op_type == TYPE_OP1D)
        expect = EXPECT_DST1;
    else if (op_type == TYPE_OP2)
        expect = EXPECT_DST;
    else if (op_type == TYPE_OP2XX)
        expect = EXPECT_OP2XX;
    else if (op_type == TYPE_REPXX)
        expect = EXPECT_REPXX;
    else if (op_type == TYPE_OP2B)
        expect = EXPECT_OP2B;
    else if (op_type == TYPE_OP3AX)
        expect = EXPECT_OP3AX;
    else if (op_type == TYPE_OP2PX)
        expect = EXPECT_OP2PX;
    else if (op_type == TYPE_OP2AX)
        expect = EXPECT_OP2AX;
    else if (op_type == TYPE_OP2EX)
        expect = EXPECT_OP2EX;
    else if (op_type == TYPE_OP2CX)
        expect = EXPECT_OP2CX;
    else if (op_type == TYPE_OP2DX)
        expect = EXPECT_OP2DX;
    else if (op_type == TYPE_OP1AX)
        expect = EXPECT_OP1AX;
    else if (op_type == TYPE_OP1B)
        expect = EXPECT_OP1B;
    else if (op_type == TYPE_OP1C)
        expect = EXPECT_OP1C;
    else
        expect = EXPECT_UNKNOWN;

    return expect;
}

int CountCommas(int i, char **tokens, int num)
{
    int count = 0;

    while (i < num)
    {
        if (!strcmp(tokens[i++], ",")) count++;
    }

    return count;
}

int CountExtras(int i, char **tokens, int num)
{
    int count = 0;

    while (i < num)
    {
        if (!strcmp(tokens[i++], "[")) count += atol(tokens[i]) - 1;
    }

    return count;
}

int CountStrings(int i, char **tokens, int num)
{
    int count = 0;

    for (;i < num; i++)
    {
        if (tokens[i][0] == '"') count += strlen(tokens[i]) - 3;
    }

    return count;
}

int CountElements(int i, char **tokens, int num)
{
    int num_elem = 1;

    if (i + 1 >= num) return 0;    
    num_elem += CountCommas(i, tokens, num);
    num_elem += CountExtras(i, tokens, num);
    num_elem += CountStrings(i, tokens, num);
    return num_elem;
}

void DumpIt(int printflag, void *ptr, int num)
{
    if (printflag != PRINT_NOCODE)
    {
        fwrite(ptr, 1, num, binfile);
        bincount += num;
    }
}

int GetData(int i, char **tokens, int num, int datasize)
{
    int count;
    int j = 0;
    int value = 0;
    int is_float = -1;
    for (;i < num; i++)
    {
        if (tokens[i][0] == '{') continue;
        if (!strcmp(tokens[i], ",")) continue;
        if (tokens[i][0] == '"')
        {
            char *ptr = tokens[i] + 1;
            while (*ptr && *ptr != '"')
            {
                value = *ptr++;
                memcpy(&databuffer[j], &value, datasize);
                j += datasize;
            }
            continue;
        }
        if (!strcmp(tokens[i], "["))
        {
            i++;
            if (EvaluateExpression(12, &i, tokens, num, &count, &is_float)) break;
            i++;
            count--;
        }
        else
        {
            count = 1;
            if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
        }
        while (count-- > 0)
        {
            memcpy(&databuffer[j], &value, datasize);
            j += datasize;
        } 
    }
    return j;
}

// Handle source for instructions such as djnz d,s/#
int ProcessRsrc(int *pi, char **tokens, int num, int *popcode, int *pexpect)
{
    int value;
    int i = *pi;
    int rflag = 0;
    int is_float = -1;

    i++;

    if (!strcmp(tokens[i], "#"))
    {
        i++;
        rflag = 1;
        *popcode |= I_BIT;
    }

    if (EvaluateExpression(12, &i, tokens, num, &value, &is_float))
    {
        *pi = i;
        return 1;
    }

    if (rflag)
    {
        if (hubmode)
        {
            value -= hub_addr + 4;
            value >>= 2;
        }
        else
            value -= (cog_addr >> 2) + 1;
    }
    *popcode |= (value & 511);

    *pi = i;
    return ProcessWx(pi, tokens, num, popcode, pexpect);
}

// Handle source for OP1S or OP2 instruction, such as jmp s or add d,s/#n
int ProcessSrc(int *pi, char **tokens, int num, int *popcode, int *pexpect, char *buffer2)
{
    int value;
    char *name;
    int extended = 0;
    int is_float = -1;

    (*pi)++;
    name = tokens[*pi];
    if (!strcmp(name, "#") || !strcmp(name, "##"))
    {
        (*pi)++;
        *popcode |= I_BIT;
        extended = !strcmp(name, "##");
    }
    if (EvaluateExpression(12, pi, tokens, num, &value, &is_float)) return 1;
    if (extended) GenerateAugx(*popcode, value, 0);
    *popcode |= (value & 511);
    return ProcessWx(pi, tokens, num, popcode, pexpect);
}

// Handle source rdxxxx and wrxxxx
int ProcessPointerSrc(int *pi, char **tokens, int num, int *popcode, int *pexpect)
{
    int value;
    int i = (*pi) + 1;

    if ((*pi = i) >= num)
    {
        PrintError("ERROR: Unexpected EOL\n", 0, 0);
        return 1;
    }
    value = EncodeAddressField(pi, tokens, num, 2, *popcode, 0);
    if (value < 0) return 1;
    if (value & 0x200) *popcode |= I_BIT;
    *popcode |= (value & 0x1ff);
    return ProcessWx(pi, tokens, num, popcode, pexpect);
}

// Parse a line in the DAT section
void ParseDat(int pass, char *buffer2, char **tokens, int num)
{
    int i;
    SymbolT *s;
    int index;
    int value;
    int rflag = 0;
    int opindex = 0;
    int is_float = -1;
    int data_size = 4;
    int cog_incr = 0;
    int hub_incr = 0;
    int printflag = PRINT_CODE;
    int opcode = 0xf0000000; // Initialize to if_always
    int expect = EXPECT_ANYTHING;
    int datalen;

    // Loop over all the tokens in this line
    for (i = 0; i < num; i++)
    {
        if (i == 0 && StrCompNoCase(tokens[0], "dat")) continue;

        // Find the token is in the symbol table
        index = FindSymbol(tokens[i]);

        // Check for an undefined symbol or local label
        if (index < 0 ||
            (pass == 1 && i == 0 && tokens[i][0] == '.'))
        {
            // If first token add the symbol as an address label
            if (expect == EXPECT_ANYTHING)
	    {
                AddSymbol2(tokens[i], cog_addr >> 2, hub_addr, hubmode ? TYPE_HUB_ADDR : TYPE_COG_ADDR);
	        if (i == num - 1)
	        {
		    printflag = PRINT_NOCODE;
		    expect = EXPECT_DONE;
		    break;
	        }
	        expect = EXPECT_IF_OR_OP;
	        continue;
	    }

            // If pass 1 replace symbol with zero since it may not be defined yet
	    else if (pass == 1)
	    {
	        index = 0;
	        SymbolTable[0].value = 0;
	    }

            // Otherwise, this is an undefined symbol in pass 2
            else
	    {
                PrintUnexpected(1, tokens[i]);
	        printflag = PRINT_NOCODE;
	        break;
	    }
        }

        // Get a pointer to the symbol in the table
        s = &SymbolTable[index];

        // Check if a label and process it.  If not a label change
        // expect to EXPECT_IF_OR_OP and continue processing.
        if (expect == EXPECT_ANYTHING)
        {
            // Check if this is an address label
	    if (s->type == TYPE_HUB_ADDR || s->type == TYPE_COG_ADDR)
	    {
	        if (pass == 2)
	        {
		    if (tokens[i][0] != '.') PurgeLocalLabels(index);
		    if (i == num - 1)
		    {
		        expect = EXPECT_DONE;
		        printflag = PRINT_NOCODE;
		        break;
		    }
		    continue;
	        }

                // If this is pass 1 then this label is already defined
	        printf("label %s is already defined\n", tokens[i]);
	        break;
	    }
	    expect = EXPECT_IF_OR_OP;
        }

        // Handle a conditional, opcode or psuedo-op
        if (expect == EXPECT_IF_OR_OP)
        {
            // Check for conditional
            if (s->type == TYPE_IF)
	    {
	        opcode = s->value;
	        expect = EXPECT_OP;
                cog_addr = (cog_addr + 3) & ~3;
                hub_addr = (hub_addr + 3) & ~3;
	        cog_incr = 4;
	        hub_incr = 4;
                continue;
	    }

            opindex = index;

            // Check for opcode or psuedo-op
            cog_incr = 4;
            hub_incr = 4;
            expect = GetExpect(s->type, i, num);
            if (expect != EXPECT_UNKNOWN)
            {
                // Check for NULL value and type, which is a NOP
                if (s->value || s->type)
                    opcode |= s->value;
                else
                    opcode = 0;
                cog_addr = (cog_addr + 3) & ~3;
                hub_addr = (hub_addr + 3) & ~3;
            }
            else if (s->type == TYPE_ORG)
            {
                hubmode = 0;
	        printflag = PRINT_NOCODE;
                hub_addr = (hub_addr + 3) & ~3;
	        if (i == num - 1)
	            cog_addr = 0;
	        else
	        {
                    i++;
                    if (EvaluateExpression(12, &i, tokens, num, &cog_addr, &is_float)) break;
                    cog_addr <<= 2;
                }
	        cog_incr = 0;
	        hub_incr = 0;
	        break;
            }
            else if (s->type == TYPE_ORGH)
            {
                int new_hub_addr = hub_addr;
                hubmode = 1;
	        printflag = PRINT_NOCODE;
	        if (i < num - 1)
	        {
                    i++;
                    if (EvaluateExpression(12, &i, tokens, num, &new_hub_addr, &is_float)) break;
                    if (new_hub_addr < hub_addr)
                        PrintError("ERROR: ORGH address %x less than previous address %x\n", (void *)new_hub_addr, (void *)hub_addr);
                    if (pass == 2)
                    {
                        int value = 0;
                        for (; hub_addr < new_hub_addr; hub_addr++) DumpIt(PRINT_CODE, &value, 1);
                    }
                    hub_addr = new_hub_addr;
                }
                cog_addr = 0;
	        cog_incr = 0;
	        hub_incr = 0;
	        break;
            }
            else if (s->type == TYPE_ALIGNL)
            {
                int value = 0;
                int cog_incr = hub_incr = (4 - hub_addr) & 3;
	        printflag = PRINT_NOCODE;
                expect = EXPECT_DONE;
                if (hub_incr && pass == 2) DumpIt(PRINT_CODE, &value, hub_incr);
	        break;
            }
            else if (s->type == TYPE_ALIGNW)
            {
                int value = 0;
                int cog_incr = hub_incr = hub_addr & 1;
	        printflag = PRINT_NOCODE;
                expect = EXPECT_DONE;
                if (hub_incr && pass == 2) DumpIt(PRINT_CODE, &value, hub_incr);
	        break;
            }
            else if (s->type == TYPE_RES)
            {
                hub_incr = 0;
	        printflag = PRINT_NOCODE;
	        if (i == num - 1)
	            cog_incr = 4;
	        else
	        {
                    i++;
                    if (EvaluateExpression(12, &i, tokens, num, &cog_incr, &is_float)) break;
                    cog_incr <<= 2;
	        }
                cog_addr = (cog_addr + 3) & ~3;
                hub_addr = (hub_addr + 3) & ~3;
	        break;
            }
            else if (s->type == TYPE_FIT)
            {
	        printflag = PRINT_NOCODE;
                cog_incr = 0;
                hub_incr = 0;
	        break;
            }
            else if (s->type == TYPE_LONG)
            {
                cog_incr = hub_incr = 4 * CountElements(i, tokens, num);
                if (pass == 2)
                {
                    int value;
                    datalen = GetData(i+1, tokens, num, 4);
                    if (datalen) DumpIt(printflag, databuffer, datalen);
                    PrintIt(printflag, hub_addr, cog_addr, datalen, buffer2, databuffer);
                }
                cog_addr += cog_incr;
                hub_addr += hub_incr;
                return;
            }
            else if (s->type == TYPE_WORD)
            {
                cog_incr = hub_incr = 2 * CountElements(i, tokens, num);
                if (pass == 2)
                {
                    int value;
                    datalen = GetData(i+1, tokens, num, 2);
                    if (datalen) DumpIt(printflag, databuffer, datalen);
                    PrintIt(printflag, hub_addr, cog_addr, datalen, buffer2, databuffer);
                }
                cog_addr += cog_incr;
                hub_addr += hub_incr;
                return;
            }
            else if (s->type == TYPE_BYTE)
            {
                cog_incr = hub_incr = CountElements(i, tokens, num);
                if (pass == 2)
                {
                    int value;
                    datalen = GetData(i+1, tokens, num, 1);
                    if (datalen) DumpIt(printflag, databuffer, datalen);
                    PrintIt(printflag, hub_addr, cog_addr, datalen, buffer2, databuffer);
                }
                cog_addr += cog_incr;
                hub_addr += hub_incr;
                return;
            }
            else if (s->type == TYPE_FILE)
            {
                char filename[100];
                int len, filesize = 0;
                FILE *tempfile;
                strcpy(filename, &tokens[i+1][1]);
                len = strlen(filename);
                if (len) filename[len-1] = 0;
                tempfile = fopen(filename, "rb");
                if (!tempfile)
                {
                    PrintError("ERROR: Couldn't open %s\n", filename, 0);
                }
                else
                {
                    fseek(tempfile, 0, SEEK_END);
                    filesize = ftell(tempfile);
                    fclose(tempfile);
                }
                cog_incr = hub_incr = filesize;
                if (pass == 2)
                {
                    fprintf(lstfile, "%5.5x %3.3x          %s\n", hub_addr, cog_addr >> 2, buffer2);
                    tempfile = fopen(filename, "rb");
                    if (tempfile)
                    {
                        while (1)
                        {
                            len = fread(databuffer, 1, 1000, tempfile);
                            if (len <= 0) break;
                            fwrite(databuffer, 1, len, binfile);
                            bincount += len;
                        }
                    }
                }
                cog_addr += cog_incr;
                hub_addr += hub_incr;
                return;
            }
            else
	    {
                PrintUnexpected(4, tokens[i]);
	        break;
	    }
	    if (pass == 1)
            {
                while (i < num)
                {
                    if (!strcmp(tokens[i++], "##"))
                    {
	                cog_addr += 4;
	                hub_addr += 4;
                    }
                }
                expect = EXPECT_DONE;
                break;
            }
        }

        // Handle an opcode
        else if (expect == EXPECT_OP)
        {
	    cog_incr = 4;
	    hub_incr = 4;
            expect = GetExpect(s->type, i, num);
            if (expect != EXPECT_UNKNOWN)
            {
	        // Check for NULL value, which is a NOP
                if (s->value || s->type)
	            opcode |= s->value;
	        else
	            opcode = 0;
            }
            else
	    {
                PrintUnexpected(5, tokens[i]);
	        break;
	    }
            opindex = index;
	    if (pass == 1)
            {
                while (i < num)
                {
                    if (!strcmp(tokens[i++], "##"))
                    {
	                cog_addr += 4;
	                hub_addr += 4;
                    }
                }
                expect = EXPECT_DONE;
                break;
            }
        }

        // Handle OP2 instruction, such as add d,s
        else if (expect == EXPECT_DST)
        {
            if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
	    opcode |= (value & 511) << 9;
	    i++;
            if (CheckExpected(",", i, tokens, num)) break;
            ProcessSrc(&i, tokens, num, &opcode, &expect, buffer2);
        }

        // Handle OP2XX instruction, such as not d[,s]
        else if (expect == EXPECT_OP2XX)
        {
            if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
	    opcode |= (value & 511) << 9;
            if (i >= num - 1 || strcmp(tokens[i+1], ","))
            {
	        opcode |= (value & 511);
                ProcessWx(&i, tokens, num, &opcode, &expect);
            }
            else
            {
	        i++;
                ProcessSrc(&i, tokens, num, &opcode, &expect, buffer2);
            }
        }

        // Handle REPXX instruction, such as rep d,s
        else if (expect == EXPECT_REPXX)
        {
            int rflag = !strcmp(s->name, "@");

            if (rflag || !strcmp(s->name, "#"))
            {
                i++;
                opcode |= Z_BIT;
            }
            if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
            if (rflag)
            {
                if (hubmode)
                    value = ((value - hub_addr) >> 2) - 2;
                else
                    value -= (cog_addr >> 2) + 2;
            }
	    opcode |= (value & 511) << 9;
            if (CheckExpected(",", ++i, tokens, num)) break;
            ProcessSrc(&i, tokens, num, &opcode, &expect, buffer2);
        }

        // Handle OP1D instruction, such as locknew d
        else if (expect == EXPECT_DST1)
        {
            if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
	    opcode |= (value & 511) << 9;
            ProcessWx(&i, tokens, num, &opcode, &expect);
        }

        // Handle OP0 instruction, such as ret
        else if (expect == EXPECT_OP0)
        {
            i--;
            ProcessWx(&i, tokens, num, &opcode, &expect);
        }

        // Handle OP2B instruction, such as rdfast d/#,s/#
        else if (expect == EXPECT_OP2B)
        {
	    value = EncodeAddressField(&i, tokens, num, 1, opcode, 1);
	    opcode |= (value & 0x1ff) << 9;
	    if (value & 0x200) opcode |= Z_BIT;
            if (CheckExpected(",", ++i, tokens, num)) break;
            ProcessSrc(&i, tokens, num, &opcode, &expect, buffer2);
        }

        // Handle OP3AX instruction, such as setnib D,S/#,#n
        else if (expect == EXPECT_OP3AX)
        {
	    value = EncodeAddressField(&i, tokens, num, 0, opcode, 1);
	    if (value < 0 || value > 511) break;
	    opcode |= (value & 0x1ff) << 9;
            if (CheckExpected(",", ++i, tokens, num)) break;
	    if (++i >= num) break;
	    value = EncodeAddressField(&i, tokens, num, 2, opcode, 0);
	    if (value < 0) break;
	    if (value & 0x200) opcode |= I_BIT;
	    opcode |= (value & 0x1ff);
            if (CheckExpected(",", ++i, tokens, num)) break;
            if (CheckExpected("#", ++i, tokens, num)) break;
            i++;
	    if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
	    opcode |= (value & 7) << 19;
            ProcessWx(&i, tokens, num, &opcode, &expect);
        }

        // Handle OP2PX instruction, such as rdlong D,S/#/PTRx
        else if (expect == EXPECT_OP2PX)
        {
	    value = EncodeAddressField(&i, tokens, num, 0, opcode, 1);
	    if (value < 0 || value > 511) break;
	    opcode |= (value & 0x1ff) << 9;
            if (CheckExpected(",", ++i, tokens, num)) break;
            ProcessPointerSrc(&i, tokens, num, &opcode, &expect);
        }

        // Handle OP2AX instruction, such as djnz D,S/#rel9
        else if (expect == EXPECT_OP2AX)
        {
            if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
	    opcode |= (value & 511) << 9;
            if (CheckExpected(",", ++i, tokens, num)) break;
            ProcessRsrc(&i, tokens, num, &opcode, &expect);
        }

        // Handle register for OP2EX instruction, such as loc reg,#abs/#rel
        else if (expect == EXPECT_OP2EX)
        {
            int is_loc = !strcmp(tokens[i-1], "loc");
            if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
            if (CheckExpected(",", ++i, tokens, num)) break;
            if (strcmp(tokens[i+1], "#") || value < 0x1f6 || value > 0x1f9)
            {
                if (is_loc)
                {
                    PrintError("ERROR: Invalid LOC instruction\n", 0, 0);
                    break;
                }
                s = &SymbolTable[opindex+1];
                opcode = s->value | (opcode & 0xf0000000);
	        opcode |= (value & 511) << 9;
                ProcessRsrc(&i, tokens, num, &opcode, &expect);
            }
            else
            {
	        opcode |= ((value - 0x1f6) & 3) << 21;
                if (CheckExpected("#", ++i, tokens, num)) break;
                ProcessBigSrc(&i, tokens, num, &opcode, &expect);
            }
        }

        // Handle OP2CX instruction, such as wrbyte D/#,S/#/PTRx
        else if (expect == EXPECT_OP2CX)
        {
	    if (strcmp(s->name, "#"))
	    {
                if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
	        opcode |= (value & 511) << 9;
                if (CheckExpected(",", ++i, tokens, num)) break;
                ProcessPointerSrc(&i, tokens, num, &opcode, &expect);
	    }
	    else
	    {
	        opcode |= Z_BIT;
	    }
        }

        // Handle OP2DX instruction, such as jp D/#,S/#rel9
        else if (expect == EXPECT_OP2DX)
        {
	    if (strcmp(s->name, "#"))
	    {
                if (EvaluateExpression(12, &i, tokens, num, &value, &is_float)) break;
	        opcode |= (value & 511) << 9;
                if (CheckExpected(",", ++i, tokens, num)) break;
                ProcessRsrc(&i, tokens, num, &opcode, &expect);
	    }
	    else
	    {
	        opcode |= Z_BIT;
	    }
        }

        // Handle OP1AX instruction, such as jmp #abs/#rel
        else if (expect == EXPECT_OP1AX)
        {
            int target_hubmode;
	    if (strcmp(s->name, "#"))
            {
                s = &SymbolTable[opindex+1];
                opcode = s->value | (opcode & 0xf0000000);
                expect = EXPECT_DST1;
                i--;
                continue;
            }
            ProcessBigSrc(&i, tokens, num, &opcode, &expect);
        }

        // Handle OP1B instruction, such as augs #23bits
        else if (expect == EXPECT_OP1B)
        {
            if (CheckExpected("#", i, tokens, num)) break;
            i++;
            EvaluateExpression(12, &i, tokens, num, &value, &is_float);
	    opcode |= (value >> 9) & 0x7fffff;
	    expect = EXPECT_DONE;
        }

        // Handle for OP1C instruction, such as cogstop d/#
        else if (expect == EXPECT_OP1C)
        {
	    value = EncodeAddressField(&i, tokens, num, 1, opcode, 1);
	    opcode |= (value & 0x1ff) << 9;
	    if (value & 0x200) opcode |= I_BIT;
            ProcessWx(&i, tokens, num, &opcode, &expect);
        }

        // Check if there are too many tokens
        else if (expect == EXPECT_DONE)
        {
            PrintUnexpected(8, tokens[i]);
	    break;
        }
    }

    // Check if there were not enough tokens
    if (expect != EXPECT_DONE && i == num)
    {
        PrintError("ERROR: Unexpected EOL\n", 0, 0);
    }

    if (pass == 2)
    {
        PrintIt(printflag, hub_addr, cog_addr, data_size, buffer2, &opcode);
        DumpIt(printflag, &opcode, 4);
    }
 
    // Update the address
    cog_addr += cog_incr;
    hub_addr += hub_incr;
    if (data_size) last_size = data_size;
}

void PrintIt(int printflag, int hub_addr, int cog_addr, int data_size, char *buffer2, void *ptr)
{
    unsigned char *value = ptr;

    if (printlevel == 0) return;

    if (printlevel >= 3)
    {
        if (hubmode)
            fprintf(lstfile, "%5.5x    ", hub_addr);
        else
            fprintf(lstfile, "%5.5x %3.3x", hub_addr, cog_addr >> 2);
    }

    if (printflag == PRINT_NOCODE)
    {
        fprintf(lstfile, "          %s\n", buffer2);
        return;
    }

    if (data_size == 0)
        fprintf(lstfile, "         ");
    else if (data_size == 1)
        fprintf(lstfile, " %2.2x      ", value[0]);
    else if (data_size == 2)
        fprintf(lstfile, " %2.2x%2.2x    ", value[1], value[0]);
    else if (data_size == 3)
        fprintf(lstfile, " %2.2x%2.2x%2.2x  ", value[2], value[1], value[0]);
    else
        fprintf(lstfile, " %2.2x%2.2x%2.2x%2.2x", value[3], value[2], value[1], value[0]);

    if (printlevel >= 2)
        fprintf(lstfile, " %s\n", buffer2);
    else
        fprintf(lstfile, "\n");

}

int CheckComment(char *buffer, int *pflag)
{
    int len = strlen(buffer);

    if (len == 0)
    {
        return *pflag;
    }
    else if (*pflag)
    {
        *pflag = (buffer[len-1] != '}');
        return 1;
    }
    else
    {
        char *ptr = SkipChars(buffer, " \t");
        if (*ptr == '{')
        {
            ptr = FindChar(ptr, '}');
            *pflag = (*ptr != '}');
        }
        return *pflag;
    }
}

int ProcessConstantLine(int currval, char **tokens, int num)
{
    int j;
    int i= 0;
    int is_float = -1;
    int commaflag = 0;

    if (num < 1) return currval;

    if (StrCompNoCase(tokens[0], "con"))
    {
        i++;
        currval = 0;
    }

    for (;i < num; i++)
    {
        if (commaflag)
        {
            commaflag = 0;
            if (CheckExpected(",", i, tokens, num)) break;
        }
        else
        {
            commaflag = 1;
            if (!strcmp(tokens[i], "#"))
            {
                if (i >= num - 1)
                {
                    fprintf(lstfile, "Expected a constant value\n");
                    break;
                }
                i++;
                EvaluateExpression(12, &i, tokens, num, &currval, &is_float);
            }
            else if (i < num - 1 && !strcmp(tokens[i+1], "="))
            {
                if (i >= num - 2)
                {
                    fprintf(lstfile, "Expected a constant value\n");
                    break;
                }
                j = i;
                i += 2;
                if (EvaluateExpression(12, &i, tokens, num, &currval, &is_float)) return currval;
                AddSymbol(tokens[j], currval++, (is_float ? TYPE_FLOAT : TYPE_CON));
            }
            else
            {
                AddSymbol(tokens[i], currval++, TYPE_CON);
            }
        }
    }

    return currval;
}

// Parse a Spin file and call ParseDat for lines in the DAT section
void Parse(int pass)
{
    int i, num;
    char *tokens[200];
    char buffer[600];
    SymbolT *s;
    int expect, index;
    int value;
    int incr;
    int mode = MODE_CON;
    int commentflag = 0;
    int currval = 0;

    cog_addr = 0;
    hub_addr = 0;
    last_size = 1;

    while (ReadString(buffer2, 300, infile, unicode))
    {
        if (CheckComment(buffer2, &commentflag))
        {
	    if (pass == 2) fprintf(lstfile, "                   %s\n", buffer2);
            continue;
        }
	num = Tokenize(buffer2, tokens, 100, buffer);
	if (num > 0) i = SearchList(SectionKeywords, tokens[0]);
	else i = -1;
	if (i >= 0 || mode != MODE_DAT || num == 0)
	{
	    if (i >= 0) mode = i;
            if (i == MODE_DAT && num > 1)
                ParseDat(pass, buffer2, tokens, num);
	    else if (pass == 2)
                fprintf(lstfile, "                   %s\n", buffer2);
	    else if (mode == MODE_CON)
                currval = ProcessConstantLine(currval, tokens, num);
	    continue;
        }
        ParseDat(pass, buffer2, tokens, num);
    }
}

void usage(void)
{
    printf("usage: qasm file\n");
    exit(1);
}

int main(int argc, char **argv)
{
    int i;
    char *ptr;
    char *infname = 0;
    char rootname[80], lstfname[80], binfname[80];

    for (i = 1; i < argc; i++)
    {
	if (argv[i][0] == '-')
	{
	    if (strncmp(argv[i], "-p", 2) == 0)
	        printlevel = argv[i][2] - '0';
	    else
	        usage();
	}
	else
           infname = argv[i];
    }

    if (!infname) usage();

    strcpy(rootname, infname);
    ptr = FindChar(rootname, '.');
    *ptr = 0;
    strcat(rootname, ".");
    strcpy(lstfname, rootname);
    strcpy(binfname, rootname);
    strcat(lstfname, "lst");
    strcat(binfname, "bin");

    infile = FileOpen(infname, "r");
    unicode = CheckForUnicode(infile);
    lstfile = FileOpen(lstfname, "w");
    binfile = FileOpen(binfname, "wb");

    ReadSymbolTable();
    Parse(1);
    Parse(2);

    i = 0;
    while (bincount & 31) DumpIt(PRINT_CODE, &i, 1);
    fclose(infile);
    fclose(lstfile);
    fclose(binfile);

    return 0;
}
